
import { Component, OnInit, OnDestroy, ViewChild, ElementRef, ChangeDetectorRef } from '@angular/core';


import { ProbeService, requesterID } from '../../../commons/service/probe.service';

import { Zone, Host, Port, Service, DiscoverHost } from '@overflow/model/discovery';
import { toMetaIPType, MetaIPTypeEnum } from '@overflow/model/meta';

import { RPCSubscriber } from '@overflow/commons/ui/decorator/RPCSubscriber';
import { DiscoveryConfigService } from '../../../commons/service/discovery-config.service';
import { Observable } from 'rxjs';

import * as d3 from 'd3';
import { Node } from '../../../commons/model/node';
import { Link } from '../../../commons/model/link';

import { RPCError } from '@overflow/rpc-js';

@Component({
  selector: 'app-pages-home',
  templateUrl: './home-page.component.html',
  styleUrls: ['./home-page.component.scss'],
})
export class HomePageComponent implements OnInit, OnDestroy {

  private discoveryTargetRef: ElementRef;

  zone: Zone;
  discoverHost: DiscoverHost;

  @ViewChild('discoveryTarget') set discoveryTarget(content: ElementRef) {
    this.discoveryTargetRef = content;
  }

  blockedPanel = false;
  displaySidebar = false;

  blockedDocument = false;

  showIntro = true;
  selectedNode = null;

  private nodes: Node[];
  private links: Link[];
  public simulation: d3.Simulation<Node, Link> | undefined;

  constructor(
    private changeDetector: ChangeDetectorRef,
    private probeService: ProbeService,
    private discoveryConfigService: DiscoveryConfigService,
  ) {
    this.nodes = [];
    this.links = [];
  }

  ngOnInit() {
    this.probeService.subscribeNotification(this);
    // this.startDiscovery("");
    this.discoveryConfigService.zone.subscribe(res => {
      this.zone = res as Zone;
    });
    this.discoveryConfigService.discoverHost.subscribe(res => {
      this.discoverHost = res as DiscoverHost;
    });
    this.discoveryConfigService.started.subscribe(res => {
      if (res) {
        this.startDiscovery();
      }
    });
  }

  ngOnDestroy(): void {
    this.probeService.unsubscribeNotification(this);
  }

  blockDocument() {
    this.blockedDocument = true;
    setTimeout(() => {
      this.blockedDocument = false;
    }, 3000);
  }

  startDiscovery() {
    // this.zone = {
    //   network: '192.168.1.0/24',
    //   iface: 'enp3s0',
    //   metaIPType: toMetaIPType(MetaIPTypeEnum.V4),
    //   address: '192.168.1.101',
    //   mac: '44:8a:5b:f1:f1:f3',
    // };

    // this.discoverHost = {
    //   metaIPType: toMetaIPType(MetaIPTypeEnum.V4),
    //   firstScanRange: '192.168.1.1',
    //   lastScanRange: '192.168.1.254',
    //   discoveryConfig: {
    //   },
    //   discoverPort: {
    //     firstScanRange: 1,
    //     lastScanRange: 65535,
    //     includeTCP: true,
    //     includeUDP: true,
    //     discoverService: {
    //     }
    //   }
    // };

    console.log('########################################################');
    console.log(this.zone);
    console.log(this.discoverHost);
    console.log('########################################################');

    this.showIntro = false;
    this.changeDetector.detectChanges();

    const width = this.discoveryTargetRef.nativeElement.clientWidth;
    const height = this.discoveryTargetRef.nativeElement.clientHeight;

    this.simulationInit();

    this.nodes = [];
    this.links = [];

    this.nodes.push(
      { id: '192.168.1.0/24', group: 'zone', target: this.zone, fx: width / 2, fy: height / 2, r: 60 },
      // { id: '192.168.1.1', group: 'host', target: { address: '192.168.1.1', zone: this.zone, deviceType: 'NAS' }, r: 40 },
      // { id: '192.168.1.1' },
      // { id: '192.168.1.2' },
    );

    this.links.push(
      // { source: this.getNode('192.168.1.0/24'), target: this.getNode('192.168.1.1') },
      // { source: this.getNode('192.168.1.0/24'), target: this.getNode('192.168.1.2') }
    );

    this.simulationRestart();

    this.probeService.send('DiscoveryService.DiscoverHost', requesterID, this.zone, this.discoverHost);

    // this.probeService.send(
    //   'DiscoveryService.DiscoverHost',
    //   requesterID,
    //   this.discoveryConfigService.getZone(),
    //   this.discoveryConfigService.getDiscoverHost()
    // );
  }

  simulationInit() {
    if (undefined !== this.simulation) {
      return;
    }

    const width = this.discoveryTargetRef.nativeElement.clientWidth;
    const height = this.discoveryTargetRef.nativeElement.clientHeight;

    const svg = d3.select(this.discoveryTargetRef.nativeElement);

    const _zoom = d3.zoom().on('zoom', () => {
      const transform = d3.event.transform;
      svg.select('g').attr('transform', 'translate(' + transform.x + ',' + transform.y + ') scale(' + transform.k + ')');
    });
    svg.call(_zoom);

    const __this = this;
    this.simulation = d3
      .forceSimulation<Node, Link>()
      .nodes(this.nodes)
      .force('charge', d3.forceManyBody().strength(-200))
      .force('link', d3.forceLink(this.links).distance(150))
      .force('center', d3.forceCenter(width / 2, height / 2))
      .force('collision', d3.forceCollide().radius(function (node: Node) {
        return node.r * 1.2;
      }))
      .on('tick', () => {
        __this.changeDetector.markForCheck();
      })
      ;
  }

  simulationRestart() {
    // Update and restart the simulation.
    const __this = this;

    function started(node: Node) {
      /** Preventing propagation of dragstart to parent elements */
      d3.event.sourceEvent.stopPropagation();

      if (!d3.event.active) {
        __this.simulation.alphaTarget(0.3).restart();
      }

      d3.event.on('drag', dragged).on('end', ended);

      function dragged() {
        if (undefined === node) {
          return;
        }

        node.fx = d3.event.x;
        node.fy = d3.event.y;
      }

      function ended() {
        if (!d3.event.active) {
          __this.simulation.alphaTarget(0);
        }

        if (undefined === node) {
          return;
        }
        node.fx = null;
        node.fy = null;
      }
    }

    d3.select(this.discoveryTargetRef.nativeElement).selectAll('.node-container').call(d3.drag().on('start', started));

    this.simulation
      .nodes(this.nodes)
      .force('link', d3.forceLink(this.links).distance(150))
      .alpha(1)
      .restart()
      ;
  }

  onTargetClick(node: Node) {
    console.log(node);

    this.displaySidebar = true;
    this.selectedNode = node;
  }

  /**
   * DiscoverHost
   */
  @RPCSubscriber({ method: 'DiscoveryService.Queueing' })
  public DiscoveryQueueing(queueingDate: Date) {
    console.log('DiscoveryQueueing', queueingDate);
  }

  /**
   * DiscoverHost
   */
  @RPCSubscriber({ method: 'DiscoveryService.QueueingFailed' })
  public DiscoveryQueueingFailed(queueingFailedDate: Date) {
    console.log('DiscoveryQueueingFailed', queueingFailedDate);
  }

  /**
   * DiscoverHost
   */
  @RPCSubscriber({ method: 'DiscoveryService.QueueingTimeout' })
  public DiscoveryQueueingTimeout(queueingTimeoutDate: Date) {
    console.log('DiscoveryQueueingTimeout', queueingTimeoutDate);
  }

  /**
   * DiscoverHost
   */
  @RPCSubscriber({ method: 'DiscoveryService.DiscoveryStart' })
  public DiscoveryStart(startDate: Date) {
    console.log('DiscoveryStart', startDate);
  }

  /**
   * DiscoverHost
   */
  @RPCSubscriber({ method: 'DiscoveryService.DiscoveryStop' })
  public DiscoveryStop(stopDate: Date) {
    console.log('DiscoveryStop', stopDate);
    this.discoveryConfigService.setStarted(false);
    this.simulation.stop();
  }

  /**
   * DiscoverHost
   */
  @RPCSubscriber({ method: 'DiscoveryService.DiscoveryError' })
  public DiscoveryError(err: RPCError) {
    console.log('DiscoveryError', err);
  }

  /**
   * DiscoveredHost
   */
  @RPCSubscriber({ method: 'DiscoveryService.DiscoveredHost' })
  public DiscoveredHost(host: Host) {
    console.log('DiscoveredHost', host);
    const hostId = `${host.address}`;
    const zoneNode = this.getNode('192.168.1.0/24');

    let hostNode = this.getNode(hostId);
    if (null !== hostNode) {
      hostNode.target = host;
    } else {
      hostNode = new Node(hostId);
      hostNode.target = host;
      hostNode.group = 'host';
      hostNode.r = 40;
      hostNode.x = zoneNode.x;
      hostNode.y = zoneNode.y;

      this.nodes.push(hostNode);
      this.links.push(new Link(zoneNode, hostNode));
    }

    this.simulationRestart();
  }

  /**
   * DiscoveredPort
   */
  @RPCSubscriber({ method: 'DiscoveryService.DiscoveredPort' })
  public DiscoveredPort(port: Port) {
    console.log('DiscoveredPort', port);
  }

  /**
   * DiscoveredService
   */
  @RPCSubscriber({ method: 'DiscoveryService.DiscoveredService' })
  public DiscoveredService(service: Service) {
    console.log('DiscoveredService', service);
    const hostId = service.port.host.address;
    const serviceId = `${service.port.host.address}-${service.port.portNumber}-${service.key}`;

    const hostNode = this.getNode(hostId);
    let serviceNode = this.getNode(serviceId);
    if (null !== serviceNode) {
      serviceNode.target = service;
    } else {
      serviceNode = new Node(serviceId);
      serviceNode.target = service;
      serviceNode.group = 'service';
      serviceNode.r = 30;
      serviceNode.x = hostNode.x;
      serviceNode.y = hostNode.y;

      this.nodes.push(serviceNode);
      this.links.push(new Link(hostNode, serviceNode));
    }

    this.simulationRestart();
  }

  private getNode(id: string): Node | null {
    let _n: Node = null;
    this.nodes.some((node): boolean => {
      if (node.id === id) {
        _n = node;
        return true;
      }
      return false;
    });
    return _n;
  }
}
